/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import java.util.HashMap;
import java.util.Map;

/**
 * Used to keep track of process (parsing or formatting) to avoid duplicates and
 * infinite loops.
 * 
 * @author ssg
 */
public class ObjectsRegistry {

    public static final String REFERENCE_DELIMITER = "#";
    public static final String REFERENCE_PREFIX = REFERENCE_DELIMITER + "REF" + REFERENCE_DELIMITER;
    // all processed objects with corresponding paths
    private Map<Integer, String> _processedObjects = new HashMap<Integer, String>();
    // all processed paths with corresponding objects
    private Map<String, Integer> _processedKeys = new HashMap<String, Integer>();
    // all objects that are referenced identified by path
    private Map<String, Object> _referencedObjects = new HashMap<String, Object>();
    // all referenced objects with path identification
    private Map<Object, String> _referenceObjects = new HashMap<Object, String>();
    private Map<Object, Object> _iid2Object = new HashMap<Object, Object>();

    /**
     * Clears cached data.
     */
    public void clear() {
        getProcessedObjects().clear();
        getProcessedKeys().clear();
        getReferencedObjects().clear();
        getReferenceObjects().clear();
        _iid2Object.clear();
    }

    /**
     * @return the _processedObjects
     */
    public Map<Integer, String> getProcessedObjects() {
        return _processedObjects;
    }

    /**
     * @return the _referencedObjects
     */
    public Map<String, Object> getReferencedObjects() {
        return _referencedObjects;
    }

    /**
     * @return the _referenceObjects
     */
    public Map<Object, String> getReferenceObjects() {
        return _referenceObjects;
    }

    public void registerProcessedObject(String objectPath, Object object) {
        //System.out.println("registerProcessedObject: " + objectPath + "; " + ((object != null) ? object.getClass().getName() : "null") + ": " + System.identityHashCode(object));
        if (objectPath != null) {
            // TODO: check if path has registered object and remove ald registrations replacing with new identityHashCode for new object...

            if (getProcessedKeys().containsKey(objectPath)) {
                Object uoid = getProcessedKeys().get(objectPath);
                if (uoid != null && uoid.equals(System.identityHashCode(object))) {
                } else {
                    // fix processed objects
                    if (_processedObjects.containsKey(System.identityHashCode(object))) {
                        _processedObjects.remove(System.identityHashCode(object));
                        //System.out.println("registerProcessedObject: " + objectPath + "; REMOVE OLD " + ((object!=null) ? object.getClass().getName() : "null"));
                    }
                    // fix referenced/reference objects
                    if (getReferenceObjects().containsKey(uoid)) {
                        String rPath = getReferenceObjects().get(uoid);
                        //getReferenceObjects().remove(uoid);
                        registerReferencedObject(rPath, object);
                        //System.out.println("    updated reference/referenced for " + rPath + ": " + uoid + " -> " + System.identityHashCode(object));
                        registerReferenceObject(object, rPath);
                    }
                }
            }
            _processedObjects.put(System.identityHashCode(object), objectPath);
            //System.out.println("registerProcessedObject: " + objectPath + "; ADD        " + ((object!=null) ? object.getClass().getName() : "null"));
            getProcessedKeys().put(objectPath, System.identityHashCode(object));
            _iid2Object.put(System.identityHashCode(object), object);
        }
    }

    public String getProcessedObjectPath(Object object) {
        if (isProcessedObject(object)) {
            return getProcessedObjects().get(System.identityHashCode(object)).toString();
        } else {
            return null;
        }
    }

    public boolean isProcessedObject(Object object) {
        return object != null && _processedObjects.containsKey(System.identityHashCode(object));
    }

    public boolean isReferencable(Object object) {
        return !Utilities.isScalarOrString(object);
    }

    public void registerReferencedObject(String objectPath, Object object) {
        // TODO: update all occurences of path
        if (objectPath != null) {
            getReferencedObjects().put(objectPath, System.identityHashCode(object));
        }
    }

    public void registerReferenceObject(Object object, String referencePath) {
        // TODO: update all occurences of path
        if (referencePath != null) {
            getReferenceObjects().put(System.identityHashCode(object), referencePath);
        }
    }

    public boolean isReference(Object object) {
        //return object instanceof String && ((String) object).startsWith(REFERENCE_PREFIX) || object != null && getReferenceObjects().containsKey(System.identityHashCode(object));
        return isReference(null, object);
    }

    public boolean isReference(String path, Object object) {
        boolean isStringRef = object instanceof String && ((String) object).startsWith(REFERENCE_PREFIX);
        if (isStringRef) {
            return true;
        }
        boolean isReferenceObject = object != null && getReferenceObjects().containsKey(System.identityHashCode(object));
        if (isReferenceObject) {
            if (path == null) {
                return true;
            }
            String refKey = getReferenceObjects().get(System.identityHashCode(object));
            if (!path.equals(refKey)) {
                return true;
            }
        }
        return false;
    }

    public Object resolveReference(String path, Object reference) {
        if (isReference(path, reference)) {// && isProcessedObject(reference)) {
            //System.out.println("resolveReference: [" + (reference instanceof String && ((String) reference).startsWith(REFERENCE_PREFIX)) + ", " + isProcessedObject(reference) + "] for " + reference);
            if (reference instanceof String) {
                // 0-""#1-"REF"#2-"className"#3-"path"
                String[] values = ((String) reference).split(REFERENCE_DELIMITER);
                Object obj = getReferencedObjects().get(values[3]);
                if (obj == null) {
                    obj = getProcessedKeys().get(values[3]);
                    if (getProcessedKeys().containsKey(values[3])) {
                        obj = _iid2Object.get(obj);
                        registerReferencedObject(values[3], obj);
                        registerReferenceObject(obj, values[3]);
                        getProcessedKeys().put(path, System.identityHashCode(obj));
                        //System.out.println("    registerResolvedReference: " + path + " -> " + values[3] + "  " + System.identityHashCode(obj));
                    }
                } else {
                    obj = _iid2Object.get(obj);
                    registerReferencedObject(values[3], obj);
                    registerReferenceObject(obj, values[3]);
                    getProcessedKeys().put(path, System.identityHashCode(obj));
                    //System.out.println("    registerResolvedReference: " + path + " -> " + values[3] + "  " + System.identityHashCode(obj));
                }
                return obj;
            } else {
                String key = getReferenceObjects().get(System.identityHashCode(reference));
                Object obj = _iid2Object.get(getReferencedObjects().get(key));
                //System.out.println("    resolvedReference: recover " + path + " -> " + System.identityHashCode(reference) + " to " + System.identityHashCode(obj));
                return obj;
            }
        }
        return reference;
    }

    /**
     * @return the _processedKeys
     */
    public Map<String, Integer> getProcessedKeys() {
        return _processedKeys;
    }

    public static String buildReferenceString(Class clazz, String path) {
        return REFERENCE_PREFIX + ((clazz != null) ? clazz.getCanonicalName() : "null") + REFERENCE_DELIMITER + path;
    }
}
