/*
 * Copyright 2011 ssg.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ssg.tools.jsonxml.utils;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/**
 * Shared methods
 * 
 * @author ssg
 */
public class Utilities {

    public static class Pair<A, B> {

        private A a;
        private B b;

        public Pair(A a, B b) {
            setA(a);
            setB(b);
        }

        /**
         * @return the a
         */
        public A getA() {
            return a;
        }

        /**
         * @param a the a to set
         */
        public void setA(A a) {
            this.a = a;
        }

        /**
         * @return the b
         */
        public B getB() {
            return b;
        }

        /**
         * @param b the b to set
         */
        public void setB(B b) {
            this.b = b;
        }
    }

    /**
     * converts noun from single form into multiple as follows (samples):
     *   offer 2 offers
     *   business 2 businesses
     *   binary 2 binaries
     *
     * @param key
     * @return
     */
    public static String single2multiple(String key) {
        if (key.endsWith("s") && !key.endsWith("ss")) {
            return key;
        } else if (key.endsWith("ss")) {
            return key + "es";
        } else if (key.endsWith("y")) {
            return key.substring(0, key.length() - 1) + "ies";
        } else {
            return key + "s";
        }
    }

    /**
     * restores noun from multiple into single form reversing
     * actions applied in single2multiple as follows (samples):
     *   offers 2 offer
     *   businesses 2 business
     *   binaries 2 binary
     *
     * @param key
     * @return
     */
    public static String multiple2single(String key) {
        if (key.endsWith("ies")) {
            return key.substring(0, key.length() - 3) + 'y';
        } else if (key.endsWith("ses")) {
            return key.substring(0, key.length() - 2);
        } else if (key.endsWith("s")) {
            return key.substring(0, key.length() - 1);
        } else {
            return key;
        }
    }

    public static boolean isNumericClass(Class clazz) {
        return clazz.equals(Byte.class)
                || clazz.equals(byte.class)
                || clazz.equals(Short.class)
                || clazz.equals(short.class)
                || clazz.equals(Integer.class)
                || clazz.equals(int.class)
                || clazz.equals(Long.class)
                || clazz.equals(long.class)
                || clazz.equals(Float.class)
                || clazz.equals(float.class)
                || clazz.equals(Double.class)
                || clazz.equals(double.class)
                || clazz.equals(BigDecimal.class);
    }

    public static boolean isScalarOrString(Object object) {
        if (object == null) {
            return true;
        }
        Class c = object.getClass();
        if (c.isPrimitive() || c.isEnum()) {
            return true;
        }
        if (object instanceof Boolean) {
            return true;
        }
        if (object instanceof Character) {
            return true;
        }
        if (object instanceof String) {
            return true;
        }
        if (object instanceof Date) {
            return true;
        }
        if (isNumericClass(object.getClass())) {
            return true;
        }
        return false;
    }

    public static Reader getReaderForResource(String name) throws IOException {
        try {
            // works in unit tests mode
            return new InputStreamReader(Thread.currentThread().getContextClassLoader().getResourceAsStream(name));
        } catch (NullPointerException npex) {
            // works in application test mode
            return new FileReader(new File(new File("./src/test/java"), name));
        }
    }

    /**
     * List interface implementation for arrays.
     * Supports objects and primitive arrays by using inner ArrayWrapper interface and
     * related implementations for: 
     *  - Object[]
     *  - byte[]
     *  - short[]
     *  - int[]
     *  - long[]
     *  - float[]
     *  - double[]
     *  - char[]
     *  - boolean[]
     * Any add/remove operation throws exception.
     * @param <T> 
     */
    public static class NonResizableListForArray<T> implements List<T> {

        ArrayWrapper<T> _array;
        int _minIdx;
        int _maxIdx;

        public interface ArrayWrapper<T> {

            int getLength();

            T get(int index);

            void set(int index, T e);
        }

        public class ObjectArray<T> implements ArrayWrapper<T> {

            T[] _a;

            public ObjectArray(T[] a) {
                _a = a;
            }

            public int getLength() {
                return _a.length;
            }

            public T get(int index) {
                return _a[index];
            }

            public void set(int index, T e) {
                _a[index] = e;
            }
        }

        public class ByteArray<T> implements ArrayWrapper<T> {

            byte[] _p;

            public ByteArray(byte[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Byte) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Byte) e;
            }
        }

        public class ShortArray<T> implements ArrayWrapper<T> {

            short[] _p;

            public ShortArray(short[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Short) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Short) e;
            }
        }

        public class IntArray<T> implements ArrayWrapper<T> {

            int[] _p;

            public IntArray(int[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Integer) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Integer) e;
            }
        }

        public class LongArray<T> implements ArrayWrapper<T> {

            long[] _p;

            public LongArray(long[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Long) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Long) e;
            }
        }

        public class FloatArray<T> implements ArrayWrapper<T> {

            float[] _p;

            public FloatArray(float[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Float) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Float) e;
            }
        }

        public class DoubleArray<T> implements ArrayWrapper<T> {

            double[] _p;

            public DoubleArray(double[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Double) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Double) e;
            }
        }

        public class CharArray<T> implements ArrayWrapper<T> {

            char[] _p;

            public CharArray(char[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Character) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Character) e;
            }
        }

        public class BoolArray<T> implements ArrayWrapper<T> {

            boolean[] _p;

            public BoolArray(boolean[] p) {
                _p = p;
            }

            public int getLength() {
                return _p.length;
            }

            public T get(int index) {
                return (T) (Boolean) _p[index];
            }

            public void set(int index, T e) {
                _p[index] = (Boolean) e;
            }
        }

        public NonResizableListForArray(Object array) {
            if (array == null) {
                array = new Object[0];
            }
            if (array.getClass().isArray()) {
                Class c = array.getClass();
                if (array instanceof byte[]) {
                    _array = new ByteArray((byte[]) array);
                } else if (array instanceof short[]) {
                    _array = new ShortArray((short[]) array);
                } else if (array instanceof int[]) {
                    _array = new IntArray((int[]) array);
                } else if (array instanceof long[]) {
                    _array = new LongArray((long[]) array);
                } else if (array instanceof float[]) {
                    _array = new FloatArray((float[]) array);
                } else if (array instanceof double[]) {
                    _array = new DoubleArray((double[]) array);
                } else if (array instanceof char[]) {
                    _array = new CharArray((char[]) array);
                } else if (array instanceof boolean[]) {
                    _array = new BoolArray((boolean[]) array);
                } else {
                    _array = new ObjectArray((T[]) array);
                }
                _minIdx = 0;
                _maxIdx = _array.getLength();
            }
        }

        public NonResizableListForArray(T[] array) {
            _array = new ObjectArray((T[]) array);
            _minIdx = 0;
            _maxIdx = _array.getLength();
        }

        public NonResizableListForArray(T[] array, int minIdx, int maxIdx) {
            _array = new ObjectArray((T[]) array);
            minIdx = Math.max(0, minIdx);
            _minIdx = minIdx;
            if (maxIdx < minIdx) {
                maxIdx = minIdx;
            }
            _maxIdx = Math.min(maxIdx + 1, _array.getLength());
        }

        public NonResizableListForArray(ArrayWrapper<T> array, int minIdx, int maxIdx) {
            _array = array;
            minIdx = Math.max(0, minIdx);
            _minIdx = minIdx;
            if (maxIdx < minIdx) {
                maxIdx = minIdx;
            }
            _maxIdx = Math.min(maxIdx + 1, _array.getLength());
        }

        public int size() {
            return _maxIdx - _minIdx;// _array.length;
        }

        public boolean isEmpty() {
            return size() == 0;//_array.length == 0;
        }

        public boolean contains(Object o) {
            for (int i = _minIdx; i < _maxIdx; i++) {
                if (o == null && _array.get(i) == null) {
                    return true;
                }
                if (_array.get(i).equals(o)) {
                    return true;
                }
            }
            return false;
        }

        public Iterator<T> iterator() {
            return listIterator();
        }

        public Object[] toArray() {
            return null;//_array;
        }

        public <T> T[] toArray(T[] a) {
            if (a == null) {
                return (T[]) toArray();
            }
            if (size() == a.length) {
                for (int i = 0; i < _array.getLength(); i++) {
                    a[i] = (T) _array.get(i + _minIdx);
                }
            }
            return a;
        }

        public boolean add(T e) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public boolean remove(Object o) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public boolean containsAll(Collection<?> c) {
            if (c == null || c.isEmpty()) {
                return true;
            }
            for (Object o : c) {
                boolean found = false;
                for (int i = _minIdx; i < _maxIdx; i++) {
                    if (_array.get(i) != null && _array.get(i).equals(o)) {
                        found = true;
                    }
                }
                if (!found) {
                    return false;
                }
            }
            return true;
        }

        public boolean addAll(Collection<? extends T> c) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public boolean addAll(int index, Collection<? extends T> c) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public boolean removeAll(Collection<?> c) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public boolean retainAll(Collection<?> c) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public void clear() {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public T get(int index) {
            int idx = _minIdx + index;
            return _array.get(idx);
        }

        public T set(int index, T element) {
            int idx = _minIdx + index;
            _array.set(idx, element);
            return element;
        }

        public void add(int index, T element) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public T remove(int index) {
            throw new UnsupportedOperationException("Not supported in non-resizable list.");
        }

        public int indexOf(Object o) {
            for (int i = _minIdx; i < _maxIdx; i++) {
                if (o == null && _array.get(i) == null) {
                    return i - _minIdx;
                }
                if (o.equals(_array.get(i))) {
                    return i - _minIdx;
                }
            }
            return -1;
        }

        public int lastIndexOf(Object o) {
            for (int i = _maxIdx - 1; i >= _minIdx; i++) {
                if (o == null && _array.get(i) == null) {
                    return i - _minIdx;
                }
                if (o.equals(_array.get(i))) {
                    return i - _minIdx;
                }
            }
            return -1;
        }

        public ListIterator<T> listIterator() {
            return listIterator(0);
        }

        public ListIterator<T> listIterator(int index) {
            final NonResizableListForArray _list = this;
            final int currentIdx = index + _minIdx;

            return new ListIterator<T>() {

                int _currentIdx = currentIdx;

                public boolean hasNext() {
                    //System.out.println("hasNext: "+_currentIdx+", "+_maxIdx+" = "+(_currentIdx < (_maxIdx - 1)));
                    return _currentIdx < _maxIdx;
                }

                public T next() {
                    if (hasNext()) {
                        return _array.get(_currentIdx++);
                    } else {
                        throw new UnsupportedOperationException("No next item is available.");
                    }
                }

                public boolean hasPrevious() {
                    return _currentIdx > _minIdx;
                }

                public T previous() {
                    if (hasNext()) {
                        return _array.get(--_currentIdx);
                    } else {
                        throw new UnsupportedOperationException("No previous item is available.");
                    }
                }

                public int nextIndex() {
                    return _currentIdx - _minIdx + 1;
                }

                public int previousIndex() {
                    return _currentIdx - _minIdx - 1;
                }

                public void remove() {
                    throw new UnsupportedOperationException("Not supported yet.");
                }

                public void set(T e) {
                    _list.set(_currentIdx, e);
                }

                public void add(T e) {
                    throw new UnsupportedOperationException("Not supported yet.");
                }
            };
        }

        public List<T> subList(int fromIndex, int toIndex) {
            return new NonResizableListForArray(_array, _minIdx + fromIndex, _minIdx + toIndex);
        }
    }
}
